/////////////////////////////////////
// Complementary Shaders by EminGT //
/////////////////////////////////////

//Common//
#include "/lib/common.glsl"

//////////Fragment Shader//////////Fragment Shader//////////Fragment Shader//////////
#ifdef FRAGMENT_SHADER

noperspective in vec2 texCoord;

//Pipeline Constants//
#include "/lib/pipelineSettings.glsl"

//Common Variables//
#if defined MC_ANISOTROPIC_FILTERING || COLORED_LIGHTING_INTERNAL > 0
    #define ANY_ERROR_MESSAGE
#endif

#ifdef MC_ANISOTROPIC_FILTERING
    #define OPTIFINE_AF_ERROR
#endif

#if COLORED_LIGHTING_INTERNAL > 0 && !defined IS_IRIS
    #define OPTIFINE_ACL_ERROR
#endif

#if COLORED_LIGHTING_INTERNAL > 0 && defined MC_OS_MAC
    #define APPLE_ACL_ERROR
#endif

#if COLORED_LIGHTING_INTERNAL > 0
    #define COORDINATES_ACL_ERROR
    #define SHADOWDISTANCE_ACL_ERROR
#endif

//Common Functions//
#if IMAGE_SHARPENING > 0
    vec2 viewD = 1.0 / vec2(viewWidth, viewHeight);

    vec2 sharpenOffsets[4] = vec2[4](
        vec2( viewD.x,  0.0),
        vec2( 0.0,  viewD.x),
        vec2(-viewD.x,  0.0),
        vec2( 0.0, -viewD.x)
    );

    void SharpenImage(inout vec3 color, vec2 texCoordM) {
        #ifdef TAA
            float sharpenMult = IMAGE_SHARPENING;
        #else
            float sharpenMult = IMAGE_SHARPENING * 0.5;
        #endif
        float mult = 0.0125 * sharpenMult;
        color *= 1.0 + 0.05 * sharpenMult;

        for (int i = 0; i < 4; i++) {
            color -= texture2D(colortex3, texCoordM + sharpenOffsets[i]).rgb * mult;
        }
    }
#endif

vec3 posterizeColor(vec3 color) {
    return mix(color, hsv2rgb(vec3(rgb2hsv(color).x, floor(rgb2hsv(color).yz*COLOR_LEVELS)/(COLOR_LEVELS-1.0))), POSTERIZE_STRENGTH);
}

vec3 sampleCell(sampler2D tex, vec2 origin, vec2 size, int count) {
    vec2 cellCoord = floor(origin / size);
    vec3 sum = vec3(0.0);
    float fCount = float(count);
    #ifdef UNDERWATER_DISTORTION
        if (isEyeInWater == 1){
            origin += WATER_REFRACTION_INTENSITY * 0.0005 * sin((cellCoord.x + cellCoord.y) * 25.0 + frameTimeCounter * 3.0);
        }
    #endif
    for (int i = 0; i < count * count; i++) {
        vec2 offset = vec2(mod(float(i), fCount) + 0.5, floor(float(i) / fCount) + 0.5) / fCount;
        sum += texture2D(tex, origin + size * offset).rgb;
    }
    return sum / (fCount * fCount);
}

vec2 getCellSize() {
    return vec2(float(int(ceil(256.0 / PIXELATED_SCREEN_SIZE_INTERNAL) + 1) & ~1)) / vec2(viewWidth, viewHeight);
}

vec3 createPixelation(sampler2D tex, vec2 uv, float sampleCount, vec2 cellSize) {
    vec2 cellOrigin = floor(uv / cellSize) * cellSize;
    sampleCount = max0(sampleCount) + 1.0;     
    
    vec3 pixelated = mix(
        sampleCell(tex, cellOrigin, cellSize, int(sampleCount)),
        sampleCell(tex, cellOrigin, cellSize, int(sampleCount) + 1),
        fract(sampleCount)
    );
    
    return posterizeColor(pixelated);
}

// Precomputed threshold map for dithering
const mat4x4 DITHER_THRESHOLD = mat4x4(
    0., 8., 2., 10., 
    12., 4., 14., 6.,
    3., 11., 1., 9.,
    15., 7., 13., 5.
);

vec3 ditherColor(vec3 color, vec2 uv, vec2 cellSize) {
    ivec2 gridCoord = ivec2(floor(uv / cellSize));
    
    ivec2 ditherIndex = gridCoord % 4;
    float ditherValue = (DITHER_THRESHOLD[ditherIndex.x][ditherIndex.y] / 16.0) - SCREEN_DITHER_I;
    
    float luminance = GetLuminance(color);
    float variance = length(color - luminance);
    
    float ditherSpread = SCREEN_DITHER_AMOUNT * variance;
    return color + ditherSpread * ditherValue;
}

#if USE_TEXTURE_PALETTE > 0
    const int MAX_PALETTE_SIZE = 32;
#else
    const int MAX_PALETTE_SIZE = 6;
    vec3 FIXED_COLOR_PALETTE[MAX_PALETTE_SIZE] = vec3[](
    vec3(PALETTE1R, PALETTE1G, PALETTE1B),
    vec3(PALETTE2R, PALETTE2G, PALETTE2B),
    vec3(PALETTE3R, PALETTE3G, PALETTE3B),
    vec3(PALETTE4R, PALETTE4G, PALETTE4B),
    vec3(PALETTE5R, PALETTE5G, PALETTE5B),
    vec3(PALETTE6R, PALETTE6G, PALETTE6B)
);
#endif

int getPaletteTextureSize() {
    return int(textureSize(depthtex2, 0).x);
}

void sortPaletteByLuminance(inout vec3 pal[MAX_PALETTE_SIZE], int size) {
    for (int i = 0; i < size - 1; i++) {
        for (int j = 0; j < size - i - 1; j++) {
            if (GetLuminance(pal[j]) > GetLuminance(pal[j + 1])) {
                vec3 temp = pal[j];
                pal[j] = pal[j + 1];
                pal[j + 1] = temp;
            }
        }
    }
}

vec3 samplePaletteFromTexture(inout vec3 paletteOut[MAX_PALETTE_SIZE], int size) {
    for (int x = 0; x < size; x++) {
        vec3 paletteColor = texelFetch(depthtex2, ivec2(x, 0), 0).rgb;       
        paletteOut[x] = paletteColor;
    }
    
    return paletteOut[0];
}

vec3 convertToPaletteColor(vec3 inputColor) {
    vec3 localPalette[MAX_PALETTE_SIZE];
    int paletteSize = MAX_PALETTE_SIZE;
    #if USE_TEXTURE_PALETTE > 0
        paletteSize = getPaletteTextureSize();
        if (paletteSize == viewWidth) {
                return vec3(1.0, 0.0, 0.0); // Bright red for visibility
        }
        samplePaletteFromTexture(localPalette, paletteSize);
    #else
        for (int i = 0; i < MAX_PALETTE_SIZE; i++) {
            localPalette[i] = FIXED_COLOR_PALETTE[i] / 255.0;
        }
    #endif

    sortPaletteByLuminance(localPalette, paletteSize);
    
    // Quantize the color based on the available palette levels.
    inputColor.r = floor(inputColor.r * float(paletteSize - 1) + 0.5) / float(paletteSize - 1);
    inputColor.g = floor(inputColor.g * float(paletteSize - 1) + 0.5) / float(paletteSize - 1);
    inputColor.b = floor(inputColor.b * float(paletteSize - 1) + 0.5) / float(paletteSize - 1);
    
    int paletteIndex = int(floor(GetLuminance(inputColor) * float(paletteSize)));
    paletteIndex = clamp(paletteIndex, 0, paletteSize - 1);

    return localPalette[paletteIndex];
}

//Includes//
// #if defined ANY_ERROR_MESSAGE || USE_TEXTURE_PALETTE > 0
    #include "/lib/textRendering/textRenderer.glsl"

    void beginTextM(int textSize, vec2 offset) {
        float scale = 860;
        beginText(ivec2(vec2(scale * viewWidth / viewHeight, scale) * texCoord) / textSize, ivec2(0 + offset.x, scale / textSize - offset.y));
        text.bgCol = vec4(0.0);
    }
// #endif

//Program//
void main() {
    vec2 texCoordM = texCoord;
    vec2 cellSize = getCellSize();

    vec3 color = createPixelation(colortex3, texCoordM, PIXELATED_SCREEN_SMOOTHNESS, cellSize);
    #ifdef SCREEN_DITHERING_INTERNAL
        color = ditherColor(color, texCoordM, cellSize);
    #endif
    #ifdef PALETTE_SWAP
        color = convertToPaletteColor(color);
    #endif

    #if CHROMA_ABERRATION > 0
        float scale = cellSize.x * viewWidth * 0.25;
        vec2 aberration = (texCoordM - 0.5) * (2.0 / vec2(viewWidth, viewHeight)) * scale * CHROMA_ABERRATION;
        color.rb = vec2(createPixelation(colortex3, texCoordM + aberration, PIXELATED_SCREEN_SMOOTHNESS, cellSize).r,
                        createPixelation(colortex3, texCoordM - aberration, PIXELATED_SCREEN_SMOOTHNESS, cellSize).b);
    #endif

    /*ivec2 boxOffsets[8] = ivec2[8](
        ivec2( 1, 0),
        ivec2( 0, 1),
        ivec2(-1, 0),
        ivec2( 0,-1),
        ivec2( 1, 1),
        ivec2( 1,-1),
        ivec2(-1, 1),
        ivec2(-1,-1)
    );

    for (int i = 0; i < 8; i++) {
        color = max(color, texelFetch(colortex3, texelCoord + boxOffsets[i], 0).rgb);
    }*/

    #ifdef OPTIFINE_AF_ERROR
        #include "/lib/textRendering/error_optifine_af.glsl"
    #elif defined OPTIFINE_ACL_ERROR
        #include "/lib/textRendering/error_optifine_acl.glsl"
    #elif defined APPLE_ACL_ERROR
        #include "/lib/textRendering/error_apple_acl.glsl"
    #else
        #ifdef COORDINATES_ACL_ERROR
            ivec2 absCameraPositionIntXZ = abs(cameraPositionInt.xz);
            if (max(absCameraPositionIntXZ.x, absCameraPositionIntXZ.y) > 8388550) {
                #include "/lib/textRendering/error_coordinates_acl.glsl"
            }
        #endif
        #ifdef SHADOWDISTANCE_ACL_ERROR
            if (COLORED_LIGHTING_INTERNAL > shadowDistance*2) {
                #include "/lib/textRendering/error_shadowdistance_acl.glsl"
            }
        #endif
    #endif
    #if USE_TEXTURE_PALETTE > 0
        if(getPaletteTextureSize() == viewWidth) {
            #if USE_TEXTURE_PALETTE == 1
                #define PALETTE_NUM _1
            #elif USE_TEXTURE_PALETTE == 2
                #define PALETTE_NUM _2
            #elif USE_TEXTURE_PALETTE == 3
                #define PALETTE_NUM _3
            #elif USE_TEXTURE_PALETTE == 4
                #define PALETTE_NUM _4
            #elif USE_TEXTURE_PALETTE == 5
                #define PALETTE_NUM _5
            #elif USE_TEXTURE_PALETTE == 6
                #define PALETTE_NUM _6
            #elif USE_TEXTURE_PALETTE == 7
                #define PALETTE_NUM _7
            #elif USE_TEXTURE_PALETTE == 8
                #define PALETTE_NUM _8
            #elif USE_TEXTURE_PALETTE == 9
                #define PALETTE_NUM _9
            #elif USE_TEXTURE_PALETTE == 10
                #define PALETTE_NUM _1, _0
            #endif
            beginTextM(8, vec2(15, 30));
                text.fgCol = vec4(0.0, 0.0, 0.0, 1.0);
                printString((_E, _R, _R, _O, _R));
                printLine();
            endText(color.rgb);
            beginTextM(4, vec2(30, 80));
                text.fgCol = vec4(1.0, 1.0, 1.0, 1.0);
                printString((_N, _o, _space, _p, _a, _l, _e, _t, _t, _e, PALETTE_NUM, _dot, _p, _n, _g, _space, _f, _o, _u, _n, _d, _space, _i, _n, _space, _t, _h, _e, _space, _p, _a, _l, _e, _t, _t, _e, _space, _f, _o, _l, _d, _e, _r));
                printLine();
                printString((_P, _l, _e, _a, _s, _e, _space, _c, _h, _e, _c, _k, _space, _s, _h, _a, _d, _e, _r, _s, _space, _gt, _space, _p, _a, _l, _e, _t, _t, _e, _s, _space, _gt, _space, _p, _a, _l, _e, _t, _t, _e, PALETTE_NUM, _dot, _p, _n, _g));
                printLine();
            endText(color.rgb);
        }
    #endif

    /* DRAWBUFFERS:0 */
    gl_FragData[0] = vec4(color, 1.0);
}

#endif

//////////Vertex Shader//////////Vertex Shader//////////Vertex Shader//////////
#ifdef VERTEX_SHADER

noperspective out vec2 texCoord;

//Attributes//

//Common Variables//

//Common Functions//

//Includes//

//Program//
void main() {
    gl_Position = ftransform();
    texCoord = (gl_TextureMatrix[0] * gl_MultiTexCoord0).xy;
}

#endif
